-------------------------------------------------------------------------------------
-- Derby Lua Script
-- By Bidwars
-------------------------------------------------------------------------------------

local M = {}
math.randomseed(os.time())  --This is used for Random Math.
local scenario = nil --Scenario Saved to this
--Helper for scenario lua commands
local helper = require('scenario/scenariohelper')

local lastPosition = {} --Last position of the vehicle
local vehIn = 0  --Number of vehicles still running
local vehOut = 0 --Number of vehicles counted out
local vTimeout = {} --Vehicles that have timed-out
local vehiclesStopTicks = {} --The vehicle timeout counter up to the maxStopTicks
local maxStopTicks = 700   --About 60 seconds --Max number of times the onPreRender(dt) loops before the vehicle is counted out
local Arena = 0  --Arena Selected
--AI Vehicles
local scenarioVehicles = {} --Table of vehicles in the scenario
local vehicleCountEntered = 0 --Number of AI vehicles
local vehicleConfig = 0  --Vehicle Config Selected equals the state
local vehicleClass = 0
local startvehicleConfig = 0
local startvehicleClass = 0  
local vehicleData = {}  --Vehicle data for each AI Vehicle
--Player Vehicle
local user = nil --User vehicle name
local player = {} --Player Object
--WayPoints
local wp1 = nil  
local wp2 = nil  
--Delay
local derbyTimer = 0 -- Used with delayStart to have a timer for when to resume
local delayStart = false  --Used Delay the Start of the senario until everything has loaded
local loadDone = false --Used to Start the senario when loading is done

require('mathlib')

-------------------------------------------------------------------------------------
--GETS A RANDOM VEHICLE JBEAM NAME
-------------------------------------------------------------------------------------
local function setupJBeam()
	local r = 0
	local JBeam = nil
	local compact = {"coupe","coupe2","hatch","pessima","sunburst","etkc","miramar","sbr"}
	local medium = {"legran","legran2","pessima","etk800","etki","midsize","midsize2","super"}
	local large = {"barstow","barstow2","fullsize","fullsize2","moonhawk","moonhawk2","burnside"}
	local truck = {"pickup","van","roamer"}
	local heavy	= {"semi"}
	local randomv = {"coupe","hatch","pessima","sunburst","etkc","miramar","sbr","legran","etk800","etki","midsize","super","barstow","fullsize","moonhawk","burnside","pickup","van","roamer"}
	math.random(); math.random(); math.random()
	if vehicleClass == 6 then --Random
		r = math.random(1, #randomv)
		JBeam = randomv[r]
	elseif vehicleClass == 5 then --Compact Car
		r = math.random(1, #compact)
		JBeam = compact[r]
	elseif vehicleClass == 4 then --Medium Car
		r = math.random(1, #medium)
		JBeam = medium[r]
	elseif vehicleClass == 3 then --Large Car
		r = math.random(1, #large)
		JBeam = large[r]
	elseif vehicleClass == 2 then --Truck
		r = math.random(1, #truck)
		JBeam = truck[r]
	else --Heavy
		r = 1
		JBeam = heavy[r]
	end
	return JBeam
end
-------------------------------------------------------------------------------------
--LOADS THE VEHICLE CLASS PC FILE
-------------------------------------------------------------------------------------
local function loadClass(vName)
	local JBeamN = nil
	local JBeam = setupJBeam()
	local number = tonumber(string.match(JBeam, "%d+"))
	if number == 2 then
		JBeamN = JBeam:sub(1, -2)
	else
		JBeamN = JBeam
	end
				
	TorqueScript.eval([[
		]]..vName..[[.JBeam = "]] .. JBeamN .. [[";
		]]..vName..[[.partConfig = "levels/derby/]]..vehicleConfig..[[_vehicles/derby_]]..JBeam..[[.pc";
		]]..vName..[[.requestReload();
		]])
	print("loaded "..vName.." config"..JBeamN.." "..JBeam.."")
end
-------------------------------------------------------------------------------------
--SETS THE PARTS FOR EACH vehicle
-------------------------------------------------------------------------------------
local function partsConfig(firstload)
	local rVehicles = scenetree.findClassObjects('BeamNGVehicle')
	for k, vName in pairs(rVehicles) do	
		local vObj = scenetree.findObject(vName)
		if vObj then
				if k == 1 then
					if startvehicleConfig == vehicleConfig and startvehicleClass == vehicleClass then
						if firstload == true then	
							loadClass(vName)	
						end
					else
						loadClass(vName)
					end
					user = vName
					player = vObj
				else
					loadClass(vName)
				end
		end
	end			
end
-------------------------------------------------------------------------------------
--RETURNS A TARGET FOR THE AI  *vNotToTarget - set AI to not target itself
-------------------------------------------------------------------------------------
local function findRandomTarget(vNotToTarget)
	local target = {}
	for vName, vObj in pairs(scenarioVehicles) do
		if vName ~= vNotToTarget and not vTimeout[vName] then
			if vehIn > 2 then
				if vehiclesStopTicks[vName] < 40 then
					if vName ~=	vehicleData[vNotToTarget].targetOut then table.insert(target, vName) end
				end
			else 
				table.insert(target, vName)
			end
		end
	end	
	math.random(); math.random(); math.random()
	local j = math.random(1, #target)
	return target[j]
end
-------------------------------------------------------------------------------------
--MOVE THE VEHICLES TO THEIR STARTING POSITION
-------------------------------------------------------------------------------------
local function moveVehiclesPosition(vName, vehicleOffset)
		--Get the Waypoint x, y, and z positions 
		local pos = {}
		pos.x = wp1.x + wp2.x
		pos.y = wp1.y + wp2.y 
		pos.z = wp1.z	
		--Offsets the vehicles so they don't get moved over one another
		pos.x = (pos.x + vehicleOffset) / 2
		pos.y = pos.y / 2
		if Arena == 1 then
		print("moved y position")
		pos.y = pos.y + 15
		end
		print("x "..pos.x.. " y "..pos.y.. " z " .. pos.z)
		--Command to move the vehicles
		TorqueScript.eval('if(isObject('..vName..')){'..vName..'.position =" '..pos.x..' '..pos.y..' '..pos.z..' ";}')	
end
-------------------------------------------------------------------------------------
--CREATE THE PREFAB OF THE VEHICLES FROM THE SCENARIO  --LOADS CONES
-------------------------------------------------------------------------------------
local function createVehicles(preFabName, vehCount)
		-- Create New Prefab Objects
		local v = createObject('Prefab')
		v.filename = "levels/derby/derby_vehicles_"..vehCount..".prefab"
		v.loadMode = 1 --'Manual'
		v.canSave  = true
		v.canSaveDynamicFields = true
		v:registerObject(preFabName)
		log('D', 'scenarios.spawnPrefab', 'loading prefab '..preFabName);
		v:load();
		scenetree.ScenarioObjectsGroup:addObject(v.obj)
		-- update static collision with new prefabs
		be:reloadCollision()	
end
local function preLoadVehicles()
	local preFabName = "aiVehicles"
	createVehicles(preFabName, vehicleCountEntered)
	partsConfig(false)
end
-------------------------------------------------------------------------------------
--GETS THE AI MOVING 
-------------------------------------------------------------------------------------
local function driveCar(vName,steering,throttle,brake,parkingbrake)
	--print("drive Car"..vName)
	scenarioVehicles[vName]:queueLuaCommand('input.event("steering", '..steering..', 1)')
	scenarioVehicles[vName]:queueLuaCommand('input.event("throttle", '..throttle..', 2)')
	scenarioVehicles[vName]:queueLuaCommand('input.event("brake", '..brake..', 2)')
	scenarioVehicles[vName]:queueLuaCommand('input.event("parkingbrake", '..parkingbrake..', 2)')
end
-------------------------------------------------------------------------------------
--SETUP THE AI 
-------------------------------------------------------------------------------------
local function setupVehicles(vName)
--print("setupVehicles "..vName)
	-- Set blank data for this ai vehicle
	vehicleData[vName] = {
				targetSet = false,
				targetOut = nil,
				target = nil,
				targetObj = {},
				dtCount = 0,
				backup = false,
				lastBackup = false,
				wallClose = false,
				forward = false,
				lastForward = false,
				startPos = vec3(0, 0, 0),
				aiAggression = .6,  --AI aggression max value 1
				}
	vehiclesStopTicks[vName] = 0 
	scenarioVehicles[vName]:queueLuaCommand('obj:requestReset(RESET_PHYSICS)')  
	scenarioVehicles[vName]:resetBrokenFlexMesh()  --Repair Vehicles
	helper.trackVehicle(vName, vName)  --Track the Vehicle
	helper.setAiMode(vName, 'manual')  --Needed to setup the AI to Auto Shifting
end
-------------------------------------------------------------------------------------
--LOAD vehicles
-------------------------------------------------------------------------------------
local function loadVehicles(scenario)
	local vehicleOffset = 0
	local sVehicles = scenetree.findClassObjects('BeamNGVehicle')
	log('D', 'scenario', "found " .. #sVehicles .. " vehicles")
	--local userVehicle = math.random(1, #sVehicles)
	for k, vName in pairs(sVehicles) do	
		local vObj = scenetree.findObject(vName)
			if vObj then	
				scenarioVehicles[vName] = vObj
				--This sets the vehicle offset amount for moving vehicles to the starting spot
				if k == 1 then
				--Move the player 
				moveVehiclesPosition(vName, vehicleOffset)
				else
				--print("Checking Even Number Vehicle Offset value" ..vehicleOffset )
						if k % 2 == 0 then
						vehicleOffset = vehicleOffset + 8
						else 
							vehicleOffset = -1 * vehicleOffset
						end
						moveVehiclesPosition(vName, vehicleOffset)
						vehicleOffset = math.abs(vehicleOffset)
				end
				vehIn = vehIn + 1 	
				lastPosition[vName] = vObj:getPosition()
				setupVehicles(vName)
				vehicleData[vName].startPos = vec3(lastPosition[vName])
			end
	end
end
-------------------------------------------------------------------------------------
--GET A LIST OF WAYPOINTS ON THE MAP
-------------------------------------------------------------------------------------
local function findWP()
local wpNum1 = 1
local wpNum2 = 2
	if Arena == 2 then
		wpNum1 = 3
		wpNum2 = 4
	elseif Arena == 1 then
		wpNum1 = 5
		wpNum2 = 6
	end
	wp1 = scenetree.findObject('wp_' ..wpNum1):getPosition()
	wp2 = scenetree.findObject('wp_' ..wpNum2):getPosition()
end
-------------------------------------------------------------------------------------
--GET THE NUMBER OF AI VEHICLES, ARENA SELECTED, AND STATE FROM THE demo_derby.js FILE
-------------------------------------------------------------------------------------
local function Selection(enteredCount,class,state,arena)
	print("Vehicle Count "..enteredCount)
	print("State "..state)
	print("Class "..class)
	print("Arena "..arena)
	vehicleCountEntered = tonumber(enteredCount)
	vehicleConfig = tonumber(state)	
	vehicleClass = tonumber(class)
	Arena = tonumber(arena)	
end
-------------------------------------------------------------------------------------
--RESET SCENARIO
-------------------------------------------------------------------------------------
local function reset()
	log('D', 'scenario', "Reset Called")
    lastPosition = {}
    vehiclesStopTicks = {}
	scenarioVehicles ={}
	derbyTimer = 0
	vehOut = 0
	vehIn = 0
	loadDone = false
end
-------------------------------------------------------------------------------------
--SCENARIO START
-------------------------------------------------------------------------------------
local function onScenarioChange(sc)
	scenario = sc
	if not scenario then reset() return end
	if scenario.state == 'pre-start' and delayStart == false then
		--Set the default partsconfig
		Selection(3,3,2,3)
		partsConfig(true)
		startvehicleConfig = vehicleConfig
		startvehicleClass = vehicleClass
	end
	if scenario.state == 'running' and scenario.vehicles and delayStart == false then			
		preLoadVehicles()
		findWP()
		loadVehicles(scenario)
		--delay the start of the senario so everything has a chance to load
		delayStart = true
		scenario.state = 'pre-running' --set the senario back to pre-running.  
		extensions.hook('onScenarioChange', scenario)
	end
end
-------------------------------------------------------------------------------------
--FUNCTION TO FREEZE AND UNFREEZE VEHICLES BEFORE THE START
-------------------------------------------------------------------------------------
local function freezeAll(state)
    for k, vObj in pairs(scenarioVehicles) do
        vObj:queueLuaCommand('scenario.freeze('..tostring(state) ..')')
    end
 end
 ------------------------------------------------------------------------------------
 --WHEN THE SENARIO COUNTDOWN TIMER REACHES GO!
 ------------------------------------------------------------------------------------
local function onScenarioRaceCountingDone()
	loadDone = true
	freezeAll(0)
end
-------------------------------------------------------------------------------------
-- AI CODE
-------------------------------------------------------------------------------------
local function runAI(vName)
	--print("called runAI "..vName)
	local data = vehicleData[vName]  --data table for the vehicle
	--if we don't have a target find one
	if data.targetSet ~= true then 
		data.target = findRandomTarget(vName)  --find a random target	
		if not data.target then data.target = user end
		data.targetObj = scenetree.findObject(data.target)
		data.targetSet = true
	 else
		 --Update Data		
		--map.objects returns objId, isactive, pos, vel, dirVec, dirVecUp, damage, objectCollisions
		local vdata = map.objects[map.objectNames[vName]]
		if vdata == nil then return end
		local tdata = map.objects[map.objectNames[data.target]]
		if tdata == nil then return end
		--Check if target is to close to a wall to target
		if vehIn > 2 then
			local tPos = tdata.pos
			if tPos.y >= wp2.y - 6  then
				data.wallClose = true
			elseif tPos.y <= wp1.y + 6 then
				data.wallClose = true
			elseif tPos.x <= wp2.x + 6 then
				data.wallClose = true
			elseif tPos.x >= wp1.x - 6 then
				data.wallClose = true
			else		
				data.wallClose = false
			end
			--if target is close to a wall then drive back to the starting position
			if data.wallClose == true then
				local pos = {}
				local vPos = vdata.pos
				if vPos.x >= data.startPos.x + 10 and vPos.x <= data.startPos.x - 10 then
					if vPos.y >= data.startPos.y + 10 and vPos.y <= data.startPos.y - 10 then
						data.targetSet = false
						data.targetOut = data.target
						data.target = nil
						data.targetObj = {}
						data.wallClose = false
					return
					end
				end
				tdata.pos = vec3(data.startPos.x, data.startPos.y, data.startPos.z)
			end
		end
		--Check the collision data to see if we hit our target
		local aiCollisions = vdata.objectCollisions
		local targetID = data.targetObj.obj:getID()  --get target id
		if aiCollisions ~= nil then
			for k,v in pairs (aiCollisions) do
				if k == targetID then
					if data.forward == true then
						data.lastForward = false
					end
					if data.backup == true then
						data.lastBackup = false
					end	
					break
				end
			end
		end
		--get the direction vector of the vehicle
		local dirVec = vdata.dirVec:normalized()
		--get the distance to the target
		local targetVec = (tdata.pos - vdata.pos):normalized()
		--if dirTarget is negitive then the target is behind
		local dirTarget = dirVec:dot(targetVec)
		--set the steering wheel angle to the target
		local dirDiff = math.asin(dirVec.x * targetVec.y - dirVec.y * targetVec.x)
		--set how hard the ai is pressing on the throttle or brake/reverse
		-- default aiAggression is .3
		local throttle = data.aiAggression  
		local brake = data.aiAggression
		if data.backup == true and data.lastBackup == false then 
			--if going foward backup after hitting the target -- also check to make sure we didn't just go backwards
			driveCar(vName,0, 0, brake, 0)
			if data.dtCount == 30 then
				data.targetSet = false
				data.target = nil
				data.targetObj = {}
				data.backup = false
				data.dtCount = 0 
				data.lastBackup = true
			else 
				data.dtCount = data.dtCount + 1
			end
			return
		elseif data.forward == true and data.lastForward == false then
			--if going backwards go foward after hitting the target  --also check to make sure we didn't just go foward
			driveCar(vName,dirDiff,throttle,0,0)
			if data.dtCount == 30 then
				data.targetSet = false
				data.target = nil
				data.targetObj = {}
				data.forward = false
				data.dtCount = 0 
				data.lastForward = true
			else 
				data.dtCount = data.dtCount + 1
			end
			return
		elseif dirTarget < 0 then
			--drive backwards and point steering wheel at the target -dirDiff
			driveCar(vName,-dirDiff,0,brake,0) 
			if vehiclesStopTicks[vName] > 40 and vehiclesStopTicks[vName] < 50 then
				data.forward = true
				data.lastForward = false
			else
				data.forward = true
				data.lastForward = true
			end
			return
		else 
			--drive forward and point steering wheel at the target -dirDiff
			driveCar(vName,-dirDiff,throttle,0,0) 
			if vehiclesStopTicks[vName] > 40 and vehiclesStopTicks[vName] < 50 then
				data.backup = true
				data.lastBackup = false
			else
				data.backup = true
				data.lastBackup = true
			end
		end	
	end
end
-------------------------------------------------------------------------------------------
--MAIN SCENARIO TIMER LOOP 
-------------------------------------------------------------------------------------------
local function onPreRender(dt)
	if delayStart == true then
		helper.realTimeUiDisplay("LOADING VEHICLES")
		if derbyTimer == 5 then
			freezeAll(1)
			be:enterVehicle(0, player.obj)
			TorqueScript.call('setCameraPlayer')
			scenario.focusSlot = player.obj:getID()
		end
		if derbyTimer == 20 then
			scenarios.onScenarioUIReady('play')
			derbyTimer = 0
			delayStart = false
		else
		derbyTimer = derbyTimer + 1
		end
	end
	if loadDone == false then return end
	--print("Running Timer")
	--vObj is the Vehicle Object and vName is the Name of the vehicle.
	for vName, vObj in pairs(scenarioVehicles) do
		if vTimeout[vName] == nil or vTimeout[vName] ~= vName then 
			if vehOut >= 1 and vehIn == 1 then
				--Senario is finished display win message
				local result = {msg = vName  ..' WINS!!!'}
				scenarios.finish(result)
				loadDone = false	
				break
			end
			if vName ~= user then
				--Get the ai moving 
				runAI(vName)
			end
				--Get the distance from the last position
			local distance =  math.abs((vObj:getPosition() - lastPosition[vName]):len())
			if distance > 0.4 then
				--vehicles still moving
				lastPosition[vName] = vObj:getPosition()
				vehiclesStopTicks[vName] = 0			
			else
				--Add one stop tick to the vehicles stop ticks
				vehiclesStopTicks[vName] = vehiclesStopTicks[vName] + 1
				--If they reach the max stop ticks then they are out
				if vehiclesStopTicks[vName] > maxStopTicks then	
					vehOut = vehOut + 1	
					vehIn = vehIn - 1
					vTimeout[vName] = vName
					local message = vName .. ' timed out!'
					guihooks.trigger('ScenarioFlashMessage', {{ message, 2.5 }})
					if vName ~= user then driveCar(vName,0, 0, 0, 1) end
				end
			end	
		end
	end
end
-------------------------------------------------------------------------------------
--DELETE VEHICLES PREFAB
-------------------------------------------------------------------------------------
local function deleteVehicles(objName)
		TorqueScript.eval('setCameraFree();')
		local obj = scenetree[objName]
		if obj then
			log('D', 'scenarios.spawnPrefab', 'unloading vehicle prefab '..objName)
			obj:unload()
			obj:delete()
			--Update static collision with new prefabs
			be:reloadCollision()
		end
end
-------------------------------------------------------------------------------------
--TRIGGERED ANYTIME THE SENARIO IS RESTARTED
-------------------------------------------------------------------------------------
local function onScenarioRestarted()
	local preFabName = "aiVehicles"
	deleteVehicles(preFabName)
end


M.Selection = Selection
M.onRaceTick = onRaceTick
M.onScenarioChange = onScenarioChange
M.onScenarioRaceCountingDone = onScenarioRaceCountingDone
M.onPreRender = onPreRender
M.onScenarioRestarted = onScenarioRestarted

return M